// Copyright 2013 The Chromium Authors
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifndef MINI_CHROMIUM_BASE_PROCESS_PROCESS_METRICS_H_
#define MINI_CHROMIUM_BASE_PROCESS_PROCESS_METRICS_H_

#include <stddef.h>

namespace base {

size_t GetPageSize();

}  // namespace base

#endif  // MINI_CHROMIUM_BASE_PROCESS_PROCESS_METRICS_H_
